from lxml import objectify, etree
from .namespaces import *

def xml(hi1) -> str:
    objectify.deannotate(hi1, cleanup_namespaces=True)
    E = objectify.ElementMaker(annotate=False, namespace=CORE_NS, nsmap=ns_map)
    result = E.HI1Message(hi1.Header, hi1.Payload)
    etree.cleanup_namespaces(result, top_nsmap=None, keep_ns_prefixes=[XSI_NS])
    etree.indent(result, space=" ")
    xml_bytes = etree.tostring(result,
                            pretty_print=True, 
                            xml_declaration=True, 
                            encoding="UTF-8"
                            )
    xml_str = xml_bytes.decode("UTF-8")
    # HACK I don't know why lxml makes xsi:type disappear
    xml_str = xml_str.replace("xsi:my_type", "xsi:type")
    return xml_str


import logging
from uuid import uuid4
from pathlib import Path
from sys import argv
from datetime import datetime, date

from lxml import etree, objectify

from .mappings import *

from ..hi1.generators import *
from ..hi1.helpers import *
from ..xml import xml

def add_lp_task_update(hi1, input_xml, identifier, auth_obj_id):
    ia_technical_id = input_xml.findall(f".//{{{FORM5_NS}}}SectionA//{{{EIO_NS}}}AuthorityTechnicalIdentifier")[0]
    taskObj = generate_LPTaskObject(uuid_for_task(auth_obj_id, identifier), ia_technical_id.Country, ia_technical_id.NationalId, auth_obj_id)
    generate_UPDATE_Action(taskObj, hi1)

    add_dict_child(taskObj, "DesiredStatus", "ETSI", "LDTaskDesiredStatus", "SubsequentProductionRequest", ns=TASK_NS)

    return taskObj    


def add_document_object(hi1, input_xml, form_id: str, parent_form_id):
    sp_technical_id = input_xml.findall(f".//{{{FORM5_NS}}}SectionA//{{{EIO_NS}}}AuthorityTechnicalIdentifier")[0]
    doc = generate_Form_Document(str(uuid4()), sp_technical_id.Country.text, sp_technical_id.NationalId.text, parent_form_id, "EPOC/EPOC-PR Form", "Form5")
    generate_CREATE_Action(doc, hi1)
    
def processForm5(input_xml, global_case_id: str, form_id: str, parent_form_id: str, num_of_identifiers = 1):
    hi1 = generate_120_request()

    # TODO - Form 5 doesn't contain any indication of the identifiers
    # so the RI will need to supply the number of identifiers (in order to generate the
    # object identifiers for the Tasks correctly)
    for i in range(num_of_identifiers):
        add_lp_task_update(hi1, input_xml, i+1, parent_form_id)

    add_document_object(hi1, input_xml, form_id, parent_form_id)

    return hi1


import logging
from uuid import uuid4
from pathlib import Path
from sys import argv, stdin, stderr
from datetime import datetime
from pprint import pformat

import argparse

from lxml import objectify, etree

from .hi1.generators import *
from .hi1.helpers import etsify_datetime
from .xml import xml
from .hi1.context import EpocContext

from .forms import form1, form2, form3, form5, form6

sender_cc = "XX"
sender_id = "RI_API"
receiver_cc = "XX"
receiver_id = "ServiceProviderID"

def main() -> None:
    args = argparse.ArgumentParser()
    args.add_argument("input", type=argparse.FileType("r"), default=stdin, help = "Path to input file, or '-' to read from stdin")
    args.add_argument("-v", "--verbose", action="count", default=0)
    pargs = args.parse_args(argv[1:])

    log_level = logging.WARNING
    match pargs.verbose:
        case 0:
            log_level = logging.WARNING
        case 1: 
            log_level = logging.INFO
        case _ if pargs.verbose >= 2:
            log_level = logging.DEBUG

    logging.basicConfig(stream=stderr, level=log_level)
    logging.info(f"Verbosity: {pargs.verbose}")
    logging.debug(f"Input arguments: {pargs}")

    logging.info(f"Parsing {pargs.input}")
    input_obj = objectify.parse(pargs.input, parser=None)

    _input_root = input_obj.getroot()
    ctx = EpocContext(
        input_root = _input_root,
        global_case_id = _input_root.globalCaseId.text,
        form_id = _input_root.formId,
        parent_form_id = str(_input_root.parentFormId) if hasattr(_input_root, "parentFormId") else None,
        form_obj = _input_root.form.getchildren()[0],
        ri_to_sp = True
    )

    logging.debug(f"Context:\n{pformat(ctx)}")

    match ctx.form_obj.tag.split("}")[1]:
        case "epocForm1" :      hi1 = form1.processForm1(ctx)
        case "epocPrForm2" :    hi1 = form2.processForm2(ctx)
        case "epocForm3" :      
            hi1 = form3.processForm3(ctx)
            ri_to_sp = False
        case "epocPrForm5" :    hi1 = form5.processForm5(ctx)
        case "epocPrForm6" :    hi1 = form6.processForm6(ctx)
        case _ :
            print (f"No matching form found for {ctx.form_obj.tag}")
            exit(-1)
    
    hi1.Header.SenderIdentifier.CountryCode = sender_cc if ctx.ri_to_sp else receiver_cc
    hi1.Header.SenderIdentifier.UniqueIdentifier = sender_id if ctx.ri_to_sp else receiver_id
    hi1.Header.ReceiverIdentifier.CountryCode = receiver_cc if ctx.ri_to_sp else sender_cc
    hi1.Header.ReceiverIdentifier.UniqueIdentifier = receiver_id if ctx.ri_to_sp else sender_id
    hi1.Header.Timestamp = etsify_datetime(datetime.now())
    hi1.Header.TransactionIdentifier = str(uuid4())

    print(xml(hi1))




from epoc2etsi.forms.globalcaseid import GlobalCaseID, INSTRUMENTS

from datetime import date, datetime, timedelta
from random import choice, randrange

LETTERS = list([chr(x) for x in range(ord('A'),ord('Z') + 1)])

def random_instrument():
    return choice(INSTRUMENTS)

def random_country():
    return choice(LETTERS) + choice(LETTERS)

def random_date():
    return date.fromtimestamp((datetime(1970,1,1) +  timedelta(days = randrange(47,482))).timestamp())

def random_date_seq():
    return randrange(0, GlobalCaseID.MAX_DATE_SEQ)

def random_des():
    return randrange(0, GlobalCaseID.MAX_NATIONAL_ID)

def random_tid():
    return randrange(0, GlobalCaseID.MAX_IDENTIFIER_ID)


def random_gci():
    gci = GlobalCaseID(random_instrument(),
                       random_country(), 
                       random_country(),
                       random_date(),
                       random_date_seq(),
                       random_des())
    return gci

def test_1():
    gci = GlobalCaseID("EPOC", "AA", "ZZ", date.today(), 0, 0)

def check_translation(gci : GlobalCaseID):
    u = gci.as_uuid()
    s = gci.as_gci()
    assert(GlobalCaseID.parse_from_gci(s).as_uuid() == u)
    assert(GlobalCaseID.parse_from_uuid(str(u)).as_gci() == s)

def test_instruments():
    for instrument in INSTRUMENTS:
        gci = random_gci()
        gci.instrument = instrument
        check_translation(gci)

def test_countries():
    for a in LETTERS:
        for b in LETTERS:
            gci = random_gci()
            gci.issuing_country = a + b
            check_translation(gci)
            gci.issuing_country = gci.executing_country
            gci.executing_country = a + b
            check_translation(gci)

def test_dates():
    early_date = date(1970,1,1)
    late_date = date(2100,1,1)
    gci = random_gci()
    gci.issue_date = early_date
    check_translation(gci)
    gci.issue_date = late_date
    check_translation(gci)
    for i in range(100):
        gci = random_gci()
        check_translation(gci)

def test_date_seq():
    for i in range(0, GlobalCaseID.MAX_DATE_SEQ):
        gci = random_gci()
        gci.issue_date_sequence = i
        check_translation(gci)

def test_national_id():
    gci = random_gci()
    for i in range(0, GlobalCaseID.MAX_NATIONAL_ID):
        gci.national_id = i
        check_translation(gci)

def test_identifier_id():
    gci = random_gci()
    for i in range(1, GlobalCaseID.MAX_IDENTIFIER_ID):
        gci.identifier_id = i
        check_translation(gci)



from datetime import datetime, date, time
from hashlib import sha256
from uuid import UUID

def etsify_datetime(d: datetime, microseconds = True) -> str:
    if microseconds:
        return str(d.astimezone().isoformat())
    else:
        s = str(d.astimezone().isoformat())
        return s[0:-13] + s[-6:]

def etsify_datetime_from_date(d: date) -> str:
    return etsify_datetime(datetime.combine(d, time(0,0,0)))

# TODO - casting ISO time to date puts it in local timezone
# create a function that can just modify the string

def etisfy_tel_no(s: str) -> str:
    return s.replace("(","")\
            .replace(")","")\
            .replace(" ","")\
            .replace("+","")



def uuid_for_task(auth_form_id: str, id_number: str):
    return str(UUID(bytes = sha256(f"{auth_form_id}-{id_number}".encode()).digest()[:16], version=4))



from lxml import objectify
from base64 import b64encode

from ..namespaces import *

def generate_120_request():
    root = objectify.fromstring('''<HI1Message xmlns="http://uri.etsi.org/03120/common/2019/10/Core" 
    xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" 
    xmlns:common="http://uri.etsi.org/03120/common/2016/02/Common" 
    xmlns:task="http://uri.etsi.org/03120/common/2020/09/Task" 
    xmlns:auth="http://uri.etsi.org/03120/common/2020/09/Authorisation">
    <Header>
        <SenderIdentifier>
            <CountryCode>XX</CountryCode>
            <UniqueIdentifier>UNKNOWN</UniqueIdentifier>
        </SenderIdentifier>
        <ReceiverIdentifier>
            <CountryCode>XX</CountryCode>
            <UniqueIdentifier>UNKNOWN</UniqueIdentifier>
        </ReceiverIdentifier>
        <TransactionIdentifier></TransactionIdentifier>
        <Timestamp></Timestamp>
        <Version>
            <ETSIVersion>V1.22.1</ETSIVersion>
            <NationalProfileOwner>EU</NationalProfileOwner>
            <NationalProfileVersion>v1.3.1</NationalProfileVersion>
        </Version>
    </Header>
    <Payload>
        <RequestPayload>
            <ActionRequests/>
        </RequestPayload>
    </Payload>
</HI1Message>
''', parser=None)
    return root


def generate_CREATE_Action(new_object, hi1):
    index = len(list(hi1.Payload.RequestPayload.ActionRequests.getchildren()))
    root = objectify.fromstring(f'''<ActionRequest>
    <ActionIdentifier>{index}</ActionIdentifier>
    <CREATE>
    </CREATE>
</ActionRequest>
''', parser=None)
    root.CREATE.append(new_object)
    hi1.Payload.RequestPayload.ActionRequests.append(root)
    return root

def generate_UPDATE_Action(new_object, hi1):
    index = len(list(hi1.Payload.RequestPayload.ActionRequests.getchildren()))
    root = objectify.fromstring(f'''<ActionRequest>
    <ActionIdentifier>{index}</ActionIdentifier>
    <UPDATE>
    </UPDATE>
</ActionRequest>
''', parser=None)
    root.UPDATE.append(new_object)
    hi1.Payload.RequestPayload.ActionRequests.append(root)
    return root    


def generate_HI1Object(object_id: str, country_code: str, owner_id: str,associated_object: str | None = None,  concrete_type: str | None = None):
    root = objectify.fromstring(f'''<HI1Object xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">
    <ObjectIdentifier>{object_id}</ObjectIdentifier>
    <CountryCode>{country_code}</CountryCode>
    <OwnerIdentifier>{owner_id}</OwnerIdentifier>
</HI1Object>''', parser=None)
    # HACK for some reason xsi:type disappears if we specify it as an attribute here
    # Instead, we call it xsi:my_type, which doesn't disappear (for some reason)
    # Then we do a string replacement later
    # No idea why this is. Come back to it later.
    if concrete_type is not None:
        root.set(f"{{{XSI_NS}}}my_type", concrete_type)
    if associated_object is not None:
        root.append(objectify.fromstring(f"<AssociatedObjects><AssociatedObject>{associated_object}</AssociatedObject></AssociatedObjects>", parser=None))
    return root

def generate_AuthorisationObject(object_id: str, country_code: str, owner_id: str, associated_object = None):
    return generate_HI1Object(object_id, country_code, owner_id, associated_object, "auth:AuthorisationObject")

def generate_LDTaskObject(object_id: str, country_code: str, owner_id: str, associated_object = None):
    return generate_HI1Object(object_id, country_code, owner_id, associated_object, "task:LDTaskObject")

def generate_LPTaskObject(object_id: str, country_code: str, owner_id: str, associated_object = None):
    return generate_HI1Object(object_id, country_code, owner_id, associated_object, "task:LPTaskObject")

def generate_DocumentObject(object_id: str, country_code: str, owner_id: str, associated_object: str):
    return generate_HI1Object(object_id, country_code, owner_id, associated_object, "doc:DocumentObject")

def generate_NotificationObject(object_id: str, country_code: str, owner_id: str, associated_object: str):
    return generate_HI1Object(object_id, country_code, owner_id, associated_object, "notify:NotificationObject")

def generate_PlaceholderObject(object_id: str, country_code: str, owner_id: str, associated_object: str):
    return generate_HI1Object(object_id, country_code, owner_id, associated_object, "ph:EPOCPlaceholderObject")



def generate_Form_Document(object_id: str, country_code: str, owner_id: str, associated_object: str, doc_name: str, doc_type: str):
    doc = generate_DocumentObject(object_id, country_code, owner_id, associated_object)
    add_child(doc, "DocumentName", doc_name, ns=DOC_NS)
    add_dict_child(doc, "DocumentType", "ETSI", "EPOCDocumentType", doc_type, ns=DOC_NS)
    body = add_child(doc, "DocumentBody", ns=DOC_NS)
    contents_b64 = b64encode(doc_name.encode("utf-8"))
    body.Contents = contents_b64
    body.ContentType = "text/plain"
    # TODO - come back and put a checksum in here if we want it
    return doc


def add_child(tree, tag_name, value = None, ns = None):
    if value is None:
        return add_cmplx_child(tree, tag_name, ns)
    M = objectify.ElementMaker(namespace=ns)
    new_element = M(tag_name, value)
    tree.append(new_element)
    return new_element

def add_cmplx_child(tree, tag_name, ns = None):
    new_element = objectify.SubElement(tree, (f"{{{ns}}}" if ns is not None else "") + tag_name)
    return new_element

def add_dict_child(tree, tag_name, dict_owner, dict_name, dict_value, ns = None):
    new_field = add_child(tree, tag_name, ns = ns)
    insert_DictionaryEntry(new_field, dict_owner, dict_name, dict_value)
    return new_field

def map_child(out_tree, tag_name, in_tree, xpath, ns = None, translation_func = None, process_func = None):
    in_element = in_tree.find(xpath)
    if in_element is not None:
        if process_func is not None:
            value = process_func(in_element)
        else:
            value = in_element.text
            if translation_func is not None:
                value = translation_func(value)
        return add_child(out_tree, tag_name, value = value, ns=ns)
    return None

def insert_DictionaryEntry(parent, owner: str, name: str, value: str):
    maker = objectify.ElementMaker(namespace=COMMON_NS, nsmap=ns_map)
    parent.append(maker.Owner(owner))
    parent.append(maker.Name(name))
    parent.append(maker.Value(value))

def insert_Flag(parent, flag_tag:str, owner: str, name: str, value: str):
    newFlag = objectify.Element(flag_tag, attrib=None, nsmap=None)
    insert_DictionaryEntry(newFlag, owner, name, value)
    parent.append(newFlag)

def flatten_children(tree):
    return "\n".join([str(c.text) for c in tree.getchildren()])


import logging
from uuid import uuid4
from sys import argv
from datetime import datetime, date

from .mappings import *

from ..hi1.generators import *
from ..hi1.helpers import *
from ..hi1.context import EpocContext


# ----------------------------------------------------------
# Authorisation object - clause 5.3.3
# ----------------------------------------------------------
def add_auth_object(hi1, input_xml, global_case_id, form_id):
    ia_technical_id = input_xml.findall(f".//{{{EPOC_NS}}}Addressee//{{{EIO_NS}}}AuthorityTechnicalIdentifier")[0]
    authObj = generate_AuthorisationObject(form_id, ia_technical_id.Country, ia_technical_id.NationalId)
    generate_CREATE_Action(authObj, hi1)

    # TODO - set authObject associated objects from Section D

    add_child(authObj, "AuthorisationReference", global_case_id, ns = AUTH_NS)
    add_dict_child(authObj, "AuthorisationLegalType", "ETSI", "EPOCLegalType", "EPOC", ns=AUTH_NS)

    deadline_tag = input_xml.find(f".//{{{FORM1_NS}}}DataProductionInRelationToDeadline")
    deadline_values = deadline_tag.findall(f".//{{{FORM1_NS}}}WithinTenDays") + deadline_tag.findall(f".//{{{FORM1_NS}}}Value")
    deadline_texts = [e.text for e in deadline_values]
    # TODO - The EC XSD allows both 10 day deadlines to be specified together. What does this mean?
    # TODO - decide how to handle multiple deadlines
    deadline = PRIORITY_MAPPING[deadline_texts[0]]
    add_dict_child(authObj, "AuthorisationPriority", "ETSI", "EPOCPriority", deadline, ns=AUTH_NS)

    add_dict_child(authObj, "AuthorisationDesiredStatus", "ETSI", "AuthorisationDesiredSatus", "SubmittedToCSP", ns=AUTH_NS)

    sp_concerned = input_xml.findall(f".//{{{EPOC_NS}}}Addressee//{{{EPOC_NS}}}ServiceProviderConcerned/{{{EPOC_NS}}}TechnicalIdentifier")
    if len(sp_concerned) > 0:
        sp_concerned = sp_concerned[0]
        auth_cspid = add_child(authObj, "AuthorisationCSPID", ns=AUTH_NS)
        auth_cspid = add_child(auth_cspid, "CSPID", ns=AUTH_NS)
        add_child(auth_cspid, "CountryCode", sp_concerned[f"{{{EIO_NS}}}Country"].text, ns=CORE_NS)
        add_child(auth_cspid, "UniqueIdentifier", sp_concerned[f"{{{EIO_NS}}}NationalId"].text, ns=CORE_NS)

    fill_in_approval_details(authObj, input_xml, is_validating_auth=False)
    if input_xml.find(f".//{{{FORM1_NS}}}SectionJ") is not None:
        fill_in_approval_details(authObj, input_xml, is_validating_auth=True)

    flags = add_child(authObj, "AuthorisationFlags", ns=AUTH_NS)
    emergency_flag = input_xml.find(f".//{{{FORM1_NS}}}SectionB/{{{EPOC_NS}}}EmergencyCase")
    if emergency_flag is not None and emergency_flag.text.lower() == "true":
        add_dict_child(flags, "AuthorisationFlag", "ETSI", "AuthorisationFlag", "IsEmergency", ns=AUTH_NS)
    delay_flag = input_xml.find(f".//{{{FORM1_NS}}}SectionH/{{{FORM1_NS}}}DelayConditions")
    if delay_flag is not None and len(delay_flag.getchildren()) > 0:
        add_dict_child(flags, "AuthorisationFlag", "ETSI", "EPOCAuthorisationFlag", "DelayInformingUser", ns=AUTH_NS)
    ea_notified = input_xml.find(f".//{{{FORM1_NS}}}SectionK/{{{FORM1_NS}}}EnforcingAuthority")
    if ea_notified is not None:
        add_dict_child(flags, "AuthorisationFlag", "ETSI", "EPOCAuthorisationFlag", "EnforcingAuthorityNotified", ns=AUTH_NS)
    # TODO - there is a technical ID for the enforcing authority buried in <ns9:SectionK> - need to work out how to map it

    map_child(authObj, "AuthorisationLegalEntity", input_xml, f".//{{{FORM1_NS}}}SectionB/{{{EPOC_NS}}}Addressee/{{{EPOC_NS}}}Authority/{{{EIO_NS}}}NameOfAuthority", ns=AUTH_NS)
    return authObj

# Mapping as described in table 5.3.3.5 (issuing) table 5.3.3-8 (validating)
def fill_in_approval_details(parent, tree, is_validating_auth):
    approval_details = add_child(parent, "AuthorisationApprovalDetails", ns=AUTH_NS)

    if is_validating_auth:
        input_section = tree.find(f".//{{{FORM1_NS}}}SectionJ/{{{EPOC_NS}}}Details")
    else:
        input_section = tree.find(f".//{{{EPOC_NS}}}IssuingAndContactAuthority/{{{EPOC_NS}}}IssuingAuthority")

    add_child(approval_details, "ApprovalType", "ValidatingAuthority" if is_validating_auth else "IssuingAuthority", ns=COMMON_NS)
    map_child(approval_details, "ApprovalReference", input_section, f"{{{EIO_NS}}}FileReference", ns=COMMON_NS)
    fill_in_approver_details(approval_details, tree, input_section, is_validating_auth)

    # TODO - there are some nasty edge-cases here to deal with re timezone offsets
    # since the XML appears to have a timezone offset but no time, and ETSI-fying the dates
    # will lead to setting the time to midnight. Suspect it is possible to end up pushing the
    # apparent time of signature back a day.
    if is_validating_auth:
        map_child(approval_details, "ApprovalTimestamp", tree, f".//{{{FORM1_NS}}}SectionJ/{{{EPOC_NS}}}Signature/{{{EPOC_NS}}}Date", ns=COMMON_NS, process_func=lambda x: etsify_datetime_from_date(date.fromisoformat(x.text.split("+")[0])))
    else:
        map_child(approval_details, "ApprovalTimestamp", tree, f".//{{{FORM1_NS}}}SectionI/{{{EPOC_NS}}}SignatureOfAuthority/{{{EPOC_NS}}}Date", ns=COMMON_NS, process_func=lambda x: etsify_datetime_from_date(date.fromisoformat(x.text.split("+")[0])))
        case_without_validation = tree.find(f".//{{{FORM1_NS}}}SectionI/{{{EPOC_NS}}}CaseWithoutValidation")
        if case_without_validation is not None and case_without_validation.text.lower() == "true":
            add_child(approval_details, "ApprovalIsEmergency", "true", ns=COMMON_NS)


# Mapping as described in table 5.3.3-5 (issuing) and 5.3.3-8 (validating)
def fill_in_approver_details(parent, tree, input_section, is_validating_auth):
    approver_details = add_child(parent, "ApproverDetails", ns=COMMON_NS)

    # TODO - check whether this can actually be absent (we have it as mandatory)
    map_child(approver_details, "ApproverName", input_section, f"{{{EIO_NS}}}NameOfAuthority", ns=COMMON_NS)    
    # TODO - check whether this can actually be absent (we have it as mandatory)
    approver_role = map_child(approver_details, 
                              "ApproverRole", 
                              tree, 
                              f".//{{{FORM1_NS}}}SectionJ/{{{EPOC_NS}}}Type" if is_validating_auth else f".//{{{FORM1_NS}}}SectionI/{{{EPOC_NS}}}Type", 
                              ns=COMMON_NS, 
                              translation_func = lambda v: ROLE_MAPPING[v])
    
    fill_in_approver_contact_details(approver_details, tree, is_alternate_poc=False, is_validating_auth=is_validating_auth)
    if not is_validating_auth:
        if input_section.find(f"../{{{EPOC_NS}}}ContactAuthority") is not None:
            fill_in_approver_contact_details(approver_details, tree, is_alternate_poc=True, is_validating_auth=False)


# Mapping as described in 5.3.3.7
def fill_in_approver_contact_details(parent, tree, is_validating_auth, is_alternate_poc):
    contact_details = add_child(parent, "ApproverContactDetails", ns=COMMON_NS)

    if is_validating_auth:
        input_section = tree.find(f".//{{{EPOC_NS}}}IssuingAndContactAuthority/{{{EPOC_NS}}}IssuingAuthority")
    elif is_alternate_poc:
        input_section = tree.find(f".//{{{EPOC_NS}}}IssuingAndContactAuthority/{{{EPOC_NS}}}ContactAuthority/{{{EPOC_NS}}}Authority")
    else:
        input_section = tree.find(f".//{{{FORM1_NS}}}SectionJ/{{{EPOC_NS}}}Details")

    if is_alternate_poc:
        map_child(contact_details, "Name", input_section, f"{{{EIO_NS}}}NameOfAuthority", ns=COMMON_NS)    
        add_child(contact_details, "Role", "Point of Contact", ns=COMMON_NS)
    else:
        map_child(contact_details, "Name", input_section, f"{{{EIO_NS}}}NameOfRepresentative", ns=COMMON_NS)    
        map_child(contact_details, "Role", input_section, f"{{{EIO_NS}}}PostHeld", ns=COMMON_NS)    
    map_child(contact_details, "EmailAddress", input_section, f"{{{EIO_NS}}}Email", ns=COMMON_NS)    
    map_child(contact_details, "PhoneNumber", input_section, f"{{{EIO_NS}}}TelNo", ns=COMMON_NS, translation_func=etisfy_tel_no)    
    map_child(contact_details, "FaxNumber", input_section, f"{{{EIO_NS}}}FaxNo", ns=COMMON_NS, translation_func=etisfy_tel_no)    
    map_child(contact_details, "Address", input_section, f"{{{EIO_NS}}}Address", ns=COMMON_NS, process_func=flatten_children)    
    if is_alternate_poc:
        return
    
    # TODO - make this translation complete
    languages = input_section.find(f".//{{{EIO_NS}}}LanguagesToCommunicate")
    if languages is not None:
        for language in languages.getchildren():
            if len(language.text) > 2:
                iso_id = LANGUAGE_MAPPING[language.text]
            else:
                iso_id = language.text
            add_child(contact_details, "Languages", ns=COMMON_NS).ISO639Set1LanguageIdentifier = iso_id

# Clause 5.4.2 and table 5.3.4-1
def add_ld_task_object(hi1, input_xml, identifier, auth_obj_id):
    ia_technical_id = input_xml.findall(f".//{{{EPOC_NS}}}Addressee//{{{EIO_NS}}}AuthorityTechnicalIdentifier")[0]
    new_uuid = uuid_for_task(auth_obj_id, identifier.Identifier.Id.text)
    logging.info(f"Generating LDTask for ID {identifier.Identifier.Id.text}, UUID {new_uuid}")
    taskObj = generate_LDTaskObject(new_uuid, ia_technical_id.Country, ia_technical_id.NationalId, auth_obj_id)
    generate_CREATE_Action(taskObj, hi1)

    # This is one way of creating an LDID for each identifier (there are many others)
    add_child(taskObj, "Reference", f"{ia_technical_id.Country}-{ia_technical_id.NationalId}-{taskObj.ObjectIdentifier}-{identifier.Identifier.Id}", ns=TASK_NS)
    add_dict_child(taskObj, "DesiredStatus", "ETSI", "LDTaskDesiredStatus", "AwaitingDisclosure", ns=TASK_NS)

    fill_in_request_details(taskObj, input_xml, identifier)
    fill_in_delivery_details(taskObj, input_xml, identifier)

    # Make sure the ManualInformation is at the end
    m = taskObj.find(f"{{{TASK_NS}}}ManualInformation")
    if m is not None:
        taskObj.append(m)
    return taskObj

def add_manual_handling(taskObj, source: str, info: str):
    m = taskObj.find(f"{{{TASK_NS}}}ManualInformation")
    if m is None:
        m = objectify.SubElement(taskObj, f"{{{TASK_NS}}}ManualInformation")
    current_text = m.text
    if current_text is not None and len(current_text) > 0:
        current_text += "\n\n"
    else:
        current_text = ""
    current_text += f"{source}:\n{info}"
    m._setText(current_text)

def fill_in_request_details(taskObj, input_xml, identifier):
    request_details = add_child(taskObj, "RequestDetails", ns=TASK_NS)

    subscriber_data = identifier.DataCategories.find(f"{{{EPOC_NS}}}Subscriber")
    user_data = identifier.DataCategories.find(f"{{{EPOC_NS}}}UserIdentification")
    traffic_data = identifier.DataCategories.find(f"{{{EPOC_NS}}}Traffic")
    content_data = identifier.DataCategories.find(f"{{{EPOC_NS}}}Content")
    
    if subscriber_data is not None and user_data is not None:
        add_dict_child(request_details, "Type", "ETSI", "SubscriberDataAndUserIdentifyingData", "SubscriberData", ns=TASK_NS)
    elif subscriber_data is not None:
        add_dict_child(request_details, "Type", "ETSI", "RequestType", "SubscriberData", ns=TASK_NS)
    elif user_data is not None:
        add_dict_child(request_details, "Type", "ETSI", "RequestType", "UserIdentifyingData", ns=TASK_NS)
    elif content_data is not None and traffic_data is not None:
        add_dict_child(request_details, "Type", "ETSI", "TrafficDataAndStoredContentData", "SubscriberData", ns=TASK_NS)
    elif content_data is not None:
        add_dict_child(request_details, "Type", "ETSI", "RequestType", "StoredContentData", ns=TASK_NS)
    elif traffic_data is not None:
        add_dict_child(request_details, "Type", "ETSI", "RequestType", "TrafficData", ns=TASK_NS)
    else:
        logging.error(f"No categories detected in Form 1 XML for identifier {taskObj.ObjectIdentifier}, aborting")
        exit()

    if hasattr(identifier.Identifier, "DateTimeRange"):
        request_details.StartTime = etsify_datetime(datetime.fromisoformat(identifier.Identifier.DateTimeRange.Start.text))
        request_details.EndTime = etsify_datetime(datetime.fromisoformat(identifier.Identifier.DateTimeRange.End.text))

    # TODO - work out how to get the ObservedTimes if IP details are specified
    fill_in_request_values(request_details, input_xml, identifier)

    category_selections = identifier.DataCategories.find(f".//{{{EPOC_NS}}}Option")
    if len(category_selections) > 0:
        sub_types = add_child(request_details, "Subtype", ns=TASK_NS)
        for category in category_selections:
            mapped_value = CATEGORY_MAPPING.get(category.text)
            if not mapped_value:
                logging.error(f"Couldn't find mapping for category type {category.text} in task {taskObj.ObjectIdentifier}")
            add_dict_child(sub_types, "RequestSubtype", "ETSI", "EPOCRequestSubtype", mapped_value, ns=TASK_NS)
    
    if subscriber_data is not None and subscriber_data.find(f"{{{EPOC_NS}}}OtherText") is not None:
        add_manual_handling(taskObj, "Section F, Subscriber Data", subscriber_data.OtherText.text)
    if user_data is not None and user_data.find(f"{{{EPOC_NS}}}Other") is not None:
        add_manual_handling(taskObj, "Section F, User Identification", user_data.Other.text)
    if identifier.DataCategories.find(f"{{{EPOC_NS}}}OtherInformation"):
        add_manual_handling(taskObj, "Section F, Other Information", identifier.DataCategories.OtherInformation)
    # TODO - come back and do TargetIdentifierSubtypes

def fill_in_request_values(request_details, input_xml, identifier):
    request_values = add_child(request_details, "RequestValues", ns=TASK_NS)
    request_value = add_child(request_values, "RequestValue", ns=TASK_NS)

    # TODO - Come back to this as/when we know how SP-provided types are being handles
    format_name = IDENTIFIER_TYPE_MAPPING[identifier.Identifier.Type.text]
    format_value = identifier.Identifier.Value.text
    target_format = add_child(request_value, "FormatType", ns=TASK_NS)
    target_format.FormatOwner = "ETSI"
    target_format.FormatName = format_name
    request_value.Value = format_value

def fill_in_delivery_details(taskObj, input_xml, identifier):
    delivery_details = add_child(taskObj, "DeliveryDetails", ns=TASK_NS)
    dest_authorities = input_xml.findall(f".//{{{FORM1_NS}}}ToWhomTransferTheData/{{{FORM1_NS}}}AuthorityCompetences")
    dest_authorities = set([c.text for c in dest_authorities])
    if "ISSUING_AUTHORITY" in dest_authorities:
        fill_in_delivery_destination(delivery_details, input_xml, input_xml.find(f".//{{{EPOC_NS}}}Addressee//{{{EIO_NS}}}AuthorityTechnicalIdentifier"), "IssuingAuthority")
    if "VALIDATING_AUTHORITY" in dest_authorities:
        fill_in_delivery_destination(delivery_details, input_xml, input_xml.find(f".//{{{FORM1_NS}}}SectionJ//{{{EIO_NS}}}AuthorityTechnicalIdentifier"), "ValidatingAuthority")
    if "OTHER_COMPETENT_AUTHORITY" in dest_authorities:
        # TODO - CCome back to this when it is clear how the techincal identifier will be transferred
        fill_in_delivery_destination(delivery_details, input_xml, None, "OtherCompetentAuthority")
    # TODO - come back to HandoverFormat, as this seems to just be free text

def fill_in_delivery_destination(delivery_details, input_xml, technical_id, destination_type):
        delivery_destination = add_child(delivery_details, "LDDeliveryDestination", ns=TASK_NS)
        if technical_id is not None:
            delivery_address = add_child(delivery_destination, "DeliveryAddress", ns=TASK_NS)
            endpoint_id = add_child(delivery_address, "EndpointID", ns=TASK_NS)
            add_child(endpoint_id, "CountryCode", technical_id.Country.text, ns=CORE_NS)
            add_child(endpoint_id, "UniqueIdentifier", technical_id.NationalId.text, ns=CORE_NS)
        add_dict_child(delivery_destination, "LDDeliveryProfile", "ETSI", "EPOCDeliveryProfile", destination_type, ns=TASK_NS)

def add_document_object(hi1, input_xml, form_id: str):
    ia_technical_id = input_xml.findall(f".//{{{EPOC_NS}}}Addressee//{{{EIO_NS}}}AuthorityTechnicalIdentifier")[0]
    doc = generate_Form_Document(str(uuid4()), ia_technical_id.Country.text, ia_technical_id.NationalId.text, form_id, "EPOC/EPOC-PR Form", "Form1")
    generate_CREATE_Action(doc, hi1)

def processForm1(ctx: EpocContext):
    hi1 = generate_120_request()

    auth_obj = add_auth_object(hi1, ctx.input_root, ctx.global_case_id, ctx.form_id)
    auth_obj_id = auth_obj.ObjectIdentifier

    identifiers = ctx.input_root.findall(f".//{{{FORM1_NS}}}SectionEToF/{{{EPOC_NS}}}Evidence")
    for identifier in identifiers:
        add_ld_task_object(hi1, ctx.input_root, identifier, auth_obj_id)

    add_document_object(hi1, ctx.input_root, ctx.form_id)

    return hi1


import logging
from uuid import uuid4
from pathlib import Path
from sys import argv
from datetime import datetime, date

from lxml import etree, objectify

from .mappings import *

from ..hi1.generators import *
from ..hi1.helpers import *
from ..xml import xml

def fill_in_status(parent, obj_id, status_dict, status_value, details):
    status = add_child(parent, "AssociatedObjectStatus", ns=NOTIFY_NS)
    status.AssociatedObject = obj_id
    add_dict_child(status, "Status", "ETSI", status_dict, status_value, ns=NOTIFY_NS)
    status.Details = details
    return status

def add_notification_object(hi1, input_xml, notif_id, original_auth_id):
    sp_technical_id = input_xml.findall(f".//{{{FORM3_NS}}}SectionH//{{{EIO_NS}}}AuthorityTechnicalIdentifier")[0]
    notif_obj = generate_NotificationObject(notif_id, sp_technical_id.Country.text, sp_technical_id.NationalId.text, original_auth_id)
    generate_CREATE_Action(notif_obj, hi1)

    add_child(notif_obj, "NotificationDetails", "Notification of Form 3", ns=NOTIFY_NS)
    add_dict_child(notif_obj, "NotificationType", "ETSI", "EPOCNotificationType", "DeFactoImpossibility", ns=NOTIFY_NS)
    add_child(notif_obj, "NewNotification", "true", ns=NOTIFY_NS)
    add_child(notif_obj, "NotificationTimestamp", etsify_datetime(datetime.now(), microseconds=False), ns=NOTIFY_NS)

    statuses = add_child(notif_obj, "StatusOfAssociatedObjects", ns=NOTIFY_NS)

    auth_error_str = "From Form 3 Section E (as an example of what could be done)\n"
    for elem in input_xml.find(f".//{{{FORM3_NS}}}SectionE").getchildren():
        if elem.tag == f"{{{FORM3_NS}}}NatureOfConflictingObligations":
            pass
        else:
            auth_error_str += f"{elem.tag.split("}")[1]}: {elem.text}\n"

    information_required = input_xml.find(f".//{{{FORM3_NS}}}SectionF/{{{FORM3_NS}}}InformationRequiredFromIssuingAuthority")
    if information_required is not None:
        auth_error_str += f"From Section F\nInformationRequiredFromIssuingAuthority: {information_required.text}"
    fill_in_status(statuses, original_auth_id, "AuthorisationStatus", "Invalid", auth_error_str)

    task_map = {}
    for identifier in input_xml.findall(f".//{{{FORM3_NS}}}SectionD/{{{FORM3_NS}}}NonExecutionReason"):
        task_id = uuid_for_task(original_auth_id, identifier.IdentifierId)
        task_error_str = ""
        if hasattr(identifier, "Reasons"):
            for reason in identifier.Reasons.getchildren():
                task_error_str += f"Reason: {reason.text}\n"
        if hasattr(identifier, "ExplanationOrOtherReason"):
            task_error_str += f"ExplanationOrOtherReason:{identifier.ExplanationOrOtherReason}"
        task_map[task_id] = task_error_str
    for identifier in input_xml.findall(f".//{{{FORM3_NS}}}SectionG/{{{FORM3_NS}}}PreservationOfData"):
        task_id = uuid_for_task(original_auth_id, identifier.IdentifierId)
        task_error_str = task_map.get(task_id, "")
        for elem in identifier.getchildren():
            task_error_str += f"{elem.tag.split("}")[1]}: {elem.text}\n"
    
    # TODO - Need to work out whether each task is an LPTask or an LDTask
    for task_id, error_str in task_map.items():
        fill_in_status(statuses, task_id, "LPTaskStatus", "Invalid", error_str)
    return hi1

def add_document_object(hi1, input_xml, form_id: str, parent_form_id):
    sp_technical_id = input_xml.findall(f".//{{{FORM3_NS}}}SectionH//{{{EIO_NS}}}AuthorityTechnicalIdentifier")[0]
    doc = generate_Form_Document(str(uuid4()), sp_technical_id.Country.text, sp_technical_id.NationalId.text, parent_form_id, "EPOC/EPOC-PR Form", "Form3")
    generate_CREATE_Action(doc, hi1)
    
def processForm3(input_xml, global_case_id: str, form_id: str, parent_form_id: str):
    hi1 = generate_120_request()

    add_notification_object(hi1, input_xml, form_id, parent_form_id)
    add_document_object(hi1, input_xml, form_id, parent_form_id)

    return hi1


from dataclasses import dataclass
from lxml.objectify import ObjectifiedElement

@dataclass
class EpocContext:
    input_root : ObjectifiedElement
    global_case_id : str
    form_id : str
    parent_form_id : str | None
    form_obj : ObjectifiedElement
    ri_to_sp : bool



XSI_NS      = "http://www.w3.org/2001/XMLSchema-instance"

EPOC_NS     = "http://data.europa.eu/edm/1/ns/epoc"
EIO_NS      = "http://data.europa.eu/edm/1/ns/eio"
FORM1_NS    = "http://data.europa.eu/edm/1/ns/forms/EPOC-FORM_1#"
FORM2_NS    = "http://data.europa.eu/edm/1/ns/forms/EPOC-PR-FORM-2#"
FORM3_NS    = "http://data.europa.eu/edm/1/ns/forms/EPOC-FORM-3#"
FORM5_NS    = "http://data.europa.eu/edm/1/ns/forms/EPOC-PR-FORM-5#"
FORM6_NS    = "http://data.europa.eu/edm/1/ns/forms/EPOC-PR-FORM-6#"

COMMON_NS   = "http://uri.etsi.org/03120/common/2016/02/Common"
CORE_NS     = "http://uri.etsi.org/03120/common/2019/10/Core"
DOC_NS      = "http://uri.etsi.org/03120/common/2020/09/Document"
TASK_NS     = "http://uri.etsi.org/03120/common/2020/09/Task"
AUTH_NS     = "http://uri.etsi.org/03120/common/2020/09/Authorisation"
# EF1_NS      = "http://uri.etsi.org/03120/common/2025/02/EpocForm1"
# PH_NS       = "http://uri.etsi.org/03120/common/2025/02/EPOCPlaceholder"
NOTIFY_NS   = "http://uri.etsi.org/03120/common/2016/02/Notification"

# EC_AUTH_NS  = "http://uri.etsi.org/03120/common/2025/02/EioAuthority"

ns_map = {
    None : CORE_NS,
    'auth' : AUTH_NS,
    'doc' : DOC_NS,
    'task' : TASK_NS,
    'notify' : NOTIFY_NS,
    # 'ef1' : EF1_NS,
    'xsi' : XSI_NS,
    # 'ph' : PH_NS,

    'epoc' : EPOC_NS,
    'eio' : EIO_NS,
    'epocform1' : FORM1_NS,
    'epocform2' : FORM2_NS,
    'epocform3' : FORM3_NS,
    'eopcform5' : FORM5_NS,
    'eopcform6' : FORM6_NS,
    'common' : COMMON_NS,
    # 'ecauth' : EC_AUTH_NS,
}


from datetime import date
from re import compile
from uuid import uuid4, UUID

from bitarray import bitarray
from bitarray.util import int2ba, ba2int

'''
        ------------------------------------------
        BIT LAYOUT FOR MAPPED EPOC + IDENTIFIER ID
        ------------------------------------------

        Bits 0-26 = Date
        Date is encoded as a 27-bit integer to represent an 8-digit decimal number
        to make it easy to translate from an ISO date (2025-01-01 becomes 20250101)
        This is quite an inefficient encoding, and can be optimised e.g. to use 
        integral days since 1970-01-01, if more space is needed.

        Bits 27-41 = Date sequence
        Encoded as a 14 bit integer representing a four digit decimal number as per 
        the EPOC GCI definition.

        Bits 42-45 = Instrument type
        Encoded as a 4 bit integer representing the index of the instrument type
        in an enum. If we assume this is mapping is only needed for EPOC/EPOC-PR,
        we can trim this down to a single bit if needed (EPOC = 0, EPOC-PR = 1)

        Bits 46-59 = Issuing country
        Country code encoded as a 10-bit integer, representing an ISO country code
        from "AA" to "ZZ" ("AA" = 0, "AB" = 1 ... "ZZ" = 675). This allows all
        ISO country codes to be carried. If it is OK to only carry those
        codes representing EU Member States, this can be made much smaller.

        The encoding is split, to accomodate the fixed "version" bits of UUID.
          Bits 46-49 = the first three bits of the issuing country
          Bits 48-51 = 1000 (as per RFC 9562 table 2)
          Bits 52-58 = remaining bits of the issuing country

        Bits 59-71 = Executing Country
        Same encoding as Issuing Country. Also split, this time to accomodate the 
        fixed "variant" bits of the UUID.
          Bits 59-63 = first five bits of the executing country
          Bits 64-65 = 10 (as per RFC 9562 table 1)
          Bits 66-70 = the remaining bits of the executing country

        Bit 71 = Padding
        Adding a padding bit here makes the DES and Identifier ID align on octet
        boundaries, which means they can be more easily parsed if needed.

        Bits 71-103 = DES
        Encoded as an unsigned 32-bit integer

        Bits 104-127 = Identifier ID
        Encoded as an unsigned 24-bit integer

        ----------------------------
        DEFINITION OF GLOBAL CASE ID
        ----------------------------
        From EDS-MESSAGE.XSD

        GlobalCaseId must uniquely identify Legal Case among different instances of DES in different countries.
        GlobalCaseId has format:
        INSTRUMENT-ISSUING_COUNTRY-EXECUTING_COUNTRY-ISSUE_DATE-ISSUE_DATE_SEQUENCE-DES_NATIONAL_INSTANCE_ID
        Dash is a separator.
        Example value: “EIO-AT-PL-2019-09-10-0012-1”.

        INSTRUMENT:
        - Instrument of the Legal Case. Must match the Service used.
        - example: EIO

        ISSUING_COUNTRY:
        - Legal Case issuing country code in ISO-3166-1 Alpha 2 format.
        - In case of EioAnnexASubmissionAction carried XML form must have the same Annex A -> Section A -> Issuing
        country.
        - example: AT

        EXECUTING_COUNTRY:
        - Legal Case executing country code in ISO-3166-1 Alpha 2 format.
        - In case of EioAnnexASubmissionAction carried XML form must have the same Annex A -> Section A -> Executing
        country.
        - example: PL

        ISSUE_DATE:
        - Legal Case issue date. Dash is separator of the fields also in issue date.
        - example: 2019-09-10

        ISSUE_DATE_SEQUENCE:
        - Sequential number of the issued Legal Cases on a given date by specific DES instance to a specific executing
        country.
        - Starts with 0001.
        - example: 0012

        DES_NATIONAL_INSTANCE_ID:
        - DES instance unique positive (non-zero) id number. Uniqueness has to be preserved within same country.
        Different countries may have the same national instance id’s configured.
        - value has to be in range: from 0 to 2^31-1 (ZERO to max integer value in Java)
        - value CANNOT be prefixed with zero(s)
        - example (CORRECT) : 1234
        - example (INCORRECT): 01234


        ----------------------------------
        UUIDv4 layout (RFC9562) clause 5.4
        ----------------------------------

        0                   1                   2                   3
        0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
        +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        |                           random_a                            |
        +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        |          random_a             |  ver  |       random_b        |
        +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        |var|                       random_c                            |
        +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        |                           random_c                            |
        +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

        For UUIDv3

'''

INSTRUMENTS = [
    "EIO",
    "MLA",
    "ITN",
    "EPOC",
    "EPOCPR",
    "SODA",
    "SODB",
    "SODX",
    "TOEA",
    "TOEL",
    "TOEX",
]

GCI_REGEX = compile(r"(EIO|MLA|ITN|EPOC|EPOCPR|SODA|SODB|SODX|TOEA|TOEL|TOEX)-[A-Z]{2}-[A-Z]{2}-[0-9]{4}-[0-9]{2}-[0-9]{2}-[0-9]{4}-[0-9]+")
TID_REGEX = compile(r"(EIO|MLA|ITN|EPOC|EPOCPR|SODA|SODB|SODX|TOEA|TOEL|TOEX)-[A-Z]{2}-[A-Z]{2}-[0-9]{4}-[0-9]{2}-[0-9]{2}-[0-9]{4}-[0-9]+-[0-9]+")

class GlobalCaseID:
    MAX_DATE_SEQ = 9999
    MAX_NATIONAL_ID = (2^32) - 1
    MAX_IDENTIFIER_ID = (2^24) - 1
    
    @staticmethod
    def parse_from_gci(s: str):
        if not (TID_REGEX.match(s) or GCI_REGEX.match(s)):
            raise ValueError(f"String {s} does not match GCI ot TID regex")
        parts = s.split("-")
        return GlobalCaseID(parts[0],
            parts[1],
            parts[2],
            date.fromisoformat(f"{parts[3]}-{parts[4]}-{parts[5]}"),
            int(parts[6]),
            int(parts[7]),
            int(parts[8]) if len(parts) == 9 else None)

    @staticmethod
    def parse_from_uuid(s: str):
        uuid = UUID(s)
        i = uuid.int
        bits = int2ba(i,128)
        bits = bits[0:48] + bits[52:64] + bits[66:]        
        tid = ba2int(bits[98:122])
        tid = None if tid == 0 else tid
        return GlobalCaseID(
            INSTRUMENTS[ba2int(bits[41:45])],
            GlobalCaseID.int_to_country(ba2int(bits[45:55])),
            GlobalCaseID.int_to_country(ba2int(bits[55:65])),
            GlobalCaseID.int_to_date(ba2int(bits[0:27])),
            ba2int(bits[27:41]),
            ba2int(bits[66:98]),
            tid
            )

    @staticmethod
    def a_to_z_as_int(c: str):
        if len(c) != 1:
            raise ValueError(f"Input value {c} not a single character")
        return ord(c[0]) - ord('A')

    @staticmethod
    def int_as_a_to_z(i: int):
        return chr(i + ord('A'))

    @staticmethod
    def country_to_int(c: str):
        if len(c) != 2:
            raise ValueError(f"Input value {c} not a valid ISO country code")
        i_hi = GlobalCaseID.a_to_z_as_int(c[0])
        i_lo = GlobalCaseID.a_to_z_as_int(c[1])
        return (i_hi * 26) + i_lo

    @staticmethod
    def int_to_country(i: int):
        i_hi = int(i / 26)
        i_lo = i % 26
        return GlobalCaseID.int_as_a_to_z(i_hi) + GlobalCaseID.int_as_a_to_z(i_lo)
    
    @staticmethod
    def date_to_int(d: date):
        return int(d.isoformat().replace("-",""))

    @staticmethod
    def int_to_date(i: int):
        s = str(i)
        s = f"{s[0:4]}-{s[4:6]}-{s[6:8]}"
        return date.fromisoformat(s)

    def __init__(self, instrument, issuing_country, executing_country, issue_date, issue_date_sequence, national_id, identifier_id = None):
        self.instrument = instrument
        self.issuing_country = issuing_country
        self.executing_country = executing_country
        self.issue_date = issue_date
        self.issue_date_sequence = issue_date_sequence
        self.national_id = national_id
        self.identifier_id = identifier_id

    def __repr__(self):
        return f"<GlobalCaseID {self.as_gci()}>"

    def as_gci(self):
        return f"{self.instrument}-{self.issuing_country}-{self.executing_country}-{self.issue_date.isoformat()}-{self.issue_date_sequence:04}-{self.national_id}{'-' + str(self.identifier_id) if self.identifier_id else ''}"

    def as_uuid(self):
        b_issuing_country = int2ba(GlobalCaseID.country_to_int(self.issuing_country), 10)
        b_executing_country = int2ba(GlobalCaseID.country_to_int(self.executing_country), 10)
    
        bits = (int2ba(GlobalCaseID.date_to_int(self.issue_date), 27) +               # Bits 0-26 = Date
                int2ba(self.issue_date_sequence, 14) +                                # Bits 27-41 = Date sequence
                int2ba(INSTRUMENTS.index(self.instrument),4) +                        # Bits 42-45 = Instrument type
                b_issuing_country[0:3] + int2ba(4,4) + b_issuing_country[3:] +        # Bits 46-59 = Issuing country (incl 4 fixed UUIDv4 "version" bits)
                b_executing_country[0:5] + int2ba(2,2) + b_executing_country[5:] +    # Bits 59-71 = Executing country (incl 2 fixed UUIDv4 "variant" bits)
                int2ba(0,1) +                                                         # 1 bit padding (to make DES and Identifier ID align on octet boundaries)
                int2ba(self.national_id, 32) +                                        # Bits 71-103 = DES
                int2ba(self.identifier_id if self.identifier_id else 0, 24))          # Bits 104-127 = Identifier ID

        i_uuid = ba2int(bits)

        u = UUID(int = i_uuid)
        return u

def uuid_from_global_case_id (global_case_id: str):
    return GlobalCaseID.parse_from_gci(global_case_id)

def global_case_id_from_uuid (uuid: str):
    return GlobalCaseID.parse_from_uuid(uuid)


from sys import argv

def gci2uuid () -> None:
    gci = GlobalCaseID.parse_from_gci(argv[1])
    uuid = gci.as_uuid()
    print(uuid)

def uuid2gci () -> None:
    gci = GlobalCaseID.parse_from_uuid(argv[1])
    print(gci.as_gci())



PRIORITY_MAPPING = {
    "AS_SOON_AS_POSSIBLE" : "TenDaysASAP",
    "AT_THE_END_OF_TEN_DAYS" : "TenDaysSubjectToEA",
    "IMMINENT_THREAT_TO_LIFE" : "EightHours",
    "IMMINENT_THREAT_TO_CRITICAL_INFRASTRUCTURE" : "EightHours",
}

ROLE_MAPPING = {
    "PUBLIC_PROSECUTOR" : "PublicProsecutor",
    "JUDGE_COURT_OR_INVESTIGATING_JUDGE"  : "JudgeCourtOrInvestigatingJudge",
    "JUDGE"  : "JudgeCourtOrInvestigatingJudge",
    "OTHER_COMPETENT_AUTHORITY" : "OtherCompetentAuthority",
}

LANGUAGE_MAPPING = {
    "fin" : "fi",
    "est" : "et",
    "eng" : "en",
    "dut" : "nl",
}

# TODO - work out what to do with "IP Address Range" and "IP Block"

IDENTIFIER_TYPE_MAPPING = {
    "IPV_4" : "IPV4Address",
    "IPV_6" : "IPV6Address",
    "IP_ADDRESS_RANGE" : "TBD",
    "IP_BLOCKS" : "TBD",
    "EMAIL" : "InternationalizedEmailAddress",
    "PHONE_NUMBER" : "InternationalE164",
    "IMEI_NUMBER" : "IMEI",
    "MAC_ADDRESS" : "MACAddress",
    "USER_ID" : "ServiceAccessIdentifier",
}

CATEGORY_MAPPING = {
    "USER_INFORMATION" : "UserInformation",
    "REGISTRATION_INFORMATION" : "RegistrationInformation",
    "TYPE_OF_SERVICE_INFORMATION" : "TypeOfServiceInformation",
    "PROFILE_INFORMATION" : "ProfileInformation",
    "VALIDATION_OF_USE_OF_SERVICE_INFORMATION" : "DataOnValidationOfUse",
    "DEBIT_OR_CREDIT_CARD_INFORMATION" : "DebitOrCreditCardInformation",
    "PUK_CODES_INFORMATION" : "PUKCodes",
    "OTHER" : "SubscriberDataOther",
}
